(********************************************************************************

   Service-Modul <ADTListe>.

   Version der ThiLLM-CD, berarbeitet von Michael Khn 
   (www.mksoftware.de.vu).
   Es gibt nun einen gendertem Inhaltstyp, sodass nicht nur 
   INTEGER-Werte mglich sind, sondern beliebige Typen.
   Voreingestellt ist ARRAY 10 OF CHAR.
   Eine nderung dieses Datentyps muss direkt im Modul 
   ADTListe (Datei adtliste.mod) erfolgen.
   
   Das Service-Modul muss eingebunden werden mit
   IMPORT ..., ADTListe;
   
********************************************************************************)
MODULE  ADTListe;
IMPORT D:=Display;

TYPE (** Dieser Typ kann hier (in der Datei adtliste.mod) gendert werden. *)
       TInhalt* = ARRAY 10 OF CHAR;
       
       
       pListe   = POINTER TO tListeIntern;

       tListeIntern = RECORD
                        daten     : TInhalt;
                        naechstes : pListe
                      END;
       TListe*   = RECORD
       (** Dies ist ein RECORD-Typ mit drei Zeigern, die vor dem Anwendungsprogrammierer
           versteckt werden. Es gibt je einen Zeiger zum ersten, zum letzten und zum 
           aktuellen Listenelement.*)
                     anker,
                     aktuelles, 
                     letztes   : pListe
                   END;


 (* --------------------  U N T E R P R O G R A M M E  -------------------- *)
 

  (* 1. Operation: *)

  PROCEDURE Erzeugen* ( VAR liste : TListe );
(** Es wird eine Liste erzeugt. Dies ist bei dynamischen Datenstrukturen
    immer ntig. Eine Definition mit *)
(** VAR meineListe: ADTListe.TListe; *)
(** gengt nicht. Es muss mit <Erzeugen> noch eine Instanz des abstrakten
    Datentyps kreiert werden. *)

   (*-----------------------------------------------------------------------*
    | Referenz: drei auf NIL gesetzte Zeiger                                |
    | Leistung: Erzeugen einer leeren Liste                                 |
    *-----------------------------------------------------------------------*)
  
    BEGIN
      liste.anker:= NIL;
      liste.aktuelles:= NIL;
      liste.letztes:= NIL
    END Erzeugen;


  (* 10. Operation *)

  PROCEDURE ListeLeer* ( liste : TListe ) : BOOLEAN;
(** Es wird geprft, ob die Liste leer ist.*)
   (*-----------------------------------------------------------------------*
    | bernahme: Verbund (drei Zeiger; DT: pListe)                          |
    | Referenz:  Funktionswert (DT: BOOLEAN)                                |
    | Leistung:  Diese Operation liefert den Wert TRUE, wenn die Liste leer |
    |            ist (Listenanker = NIL), sonst FALSE.                      |
    *-----------------------------------------------------------------------*)

    BEGIN
      RETURN liste.anker = NIL
    END ListeLeer;


  PROCEDURE ListenAnfang ( liste : TListe ) : BOOLEAN;
   (*-----------------------------------------------------------------------*
    | bernahme: Verbund (drei Zeiger; DT: pListe)                          |
    | Referenz:  Funktionswert (DT: BOOLEAN)                                |
    | Leistung:  Diese Operation liefert den Wert TRUE, wenn das erste Ele- |
    |            ment aktuelles Element ist. TRUE wird auch geliefert, wenn |
    |            die Liste leer ist.                                        |
    |            Sonst liefert die Operation den Wert FALSE.                |
    *-----------------------------------------------------------------------*)

    BEGIN
      IF ListeLeer(liste) THEN RETURN TRUE
        ELSE RETURN liste.aktuelles = liste.anker
      END
    END ListenAnfang;


  (* 11. Operation *)

  PROCEDURE ListenEnde* ( liste : TListe ) : BOOLEAN;
(** Diese Funktion liefert den Wert TRUE, wenn das aktuelle Element das
    letzte der Liste ist oder wenn die Liste leer ist. Ansonsten ist der
    Wert FALSE. *)
   (*-----------------------------------------------------------------------*
    | bernahme: Verbund (drei Zeiger; DT: pListe)                          |
    | Referenz:  Funktionswert (DT: BOOLEAN)                                |
    | Leistung:  Diese Operation liefert den Wert TRUE, wenn die Liste leer |
    |            ist (Listenanker = NIL) oder der Zeiger auf das aktuelle   |
    |            Element und der auf das letze Element identisch sind,      |
    |            sonst FALSE.                                               |
    *-----------------------------------------------------------------------*)

    BEGIN
      IF ListeLeer(liste) THEN RETURN TRUE
        ELSE RETURN liste.aktuelles = liste.letztes
      END
    END ListenEnde;


  PROCEDURE Davor ( liste : TListe ) : pListe;

   (*-----------------------------------------------------------------------*
    | bernahme: Verbund (drei Zeiger; DT: pListe)                          |
    | Referenz:  Funktionswert (Zeiger auf ein Listenelement; DT: pListe)   |
    | Leistung:  Diese Operation liefert einen Zeiger auf das Element vor   |
    |            dem aktuellen Element. NIL wird geliefert, wenn die Liste  |
    |            leer oder das erste Element aktuelles Element ist.         |
    *-----------------------------------------------------------------------*)

    VAR zeiger : pListe;

    BEGIN
      IF ListenAnfang(liste) THEN RETURN NIL
      ELSE      (* Die Liste enthlt mindestens  Z W E I  Elemente.         *)
                (* Das erste Element der Liste ist nicht aktuelles Element. *)
        zeiger:= liste.anker;
        WHILE zeiger^.naechstes # liste.aktuelles DO
          zeiger:= zeiger^.naechstes
        END;
        RETURN zeiger
      END
    END Davor;


  (* 2. Operation *)

  PROCEDURE EinfuegenElement* ( VAR liste : TListe; inhalt : TInhalt ); 
(** Es wird in die <liste> ein neues Element mit dem angegebenen <inhalt>
    eingefgt. *)
(** Dabei wird das neue Element <vor> das gerade aktuelle Element eingefgt. 
    Wenn die Liste zuvor leer war oder nur aus einem Element bestand, wird
    das neue Element zum ersten der Liste. *)
(** Nach dieser Operation ist das neu eingefgte Element auch das aktuelle. *)

   (*-----------------------------------------------------------------------*
    | bernahme: Verbund (drei Zeiger; DT: pListe) und Inhalt (DT: inhaltT) |
    | Referenz:  um ein Element erweiterte Liste, aktualisierte Zeiger      |
    | Leistung:  Neues Listenelement 'kreieren' (Speicherplatz vom Heap an- |
    |            fordern) und der Variable neues^.daten den Wert der Vari-  |
    |            ablen inhalt zuweisen. Falls die Liste leer ist, wird das  |
    |            neue Element als erstes Element in die Liste bernommen,   |
    |            dem Zeiger des neuen Elements NIL zugewiesen und die ande- |
    |            en Zeiger aktualisiert. Falls die Liste bereits ein Ele-   |
    |            ment enthlt, wird das neue Element am Anfang der Liste    |
    |            eingefgt und die Zeiger aktualisiert. Sollte beides nicht |
    |            zutreffen, das erste Element ist nicht aktuelles Element   |
    |            der Liste, dann wird das neue Listenelement vor dem der-   |
    |            zeit aktuellen Element in die Liste eingefgt und die Zei- |
    |            ger werden aktualisiert.                                   |
    *-----------------------------------------------------------------------*)

    VAR neues, voraktuelles : pListe;

    BEGIN
      NEW(neues);                                    (* neues Listenelement *)
      neues^.daten:= inhalt;
      IF ListeLeer(liste) THEN                       (* Die Liste ist leer. *)
        neues^.naechstes:= NIL;
        liste.anker:= neues;
        liste.aktuelles:= neues;
        liste.letztes:= neues
      ELSIF ListenAnfang(liste) THEN 
                                (* Das erste Element ist aktuelles Element. *)
        neues^.naechstes:= liste.aktuelles;
        liste.anker:= neues;
        liste.aktuelles:= neues
      ELSE                (* Das erste Element ist nicht aktuelles Element. *)
        voraktuelles:= Davor(liste);
        voraktuelles^.naechstes:= neues;
        neues^.naechstes:= liste.aktuelles;
        liste.aktuelles:= neues
      END
    END EinfuegenElement;


  (* 3. Operation *)

  PROCEDURE AnhaengenElement* ( VAR liste : TListe; inhalt : TInhalt );
(** Es wird am Ende der <liste> ein neues Element mit dem angegebenen 
    <inhalt> eingefgt. *)
(** Sollte die Liste leer sein, wird das neue Element zum ersten. *)
(** Das neue Listen-Element wird dabei auch zum aktuellen Element. *)
   (*-----------------------------------------------------------------------*
    | bernahme: Verbund (drei Zeiger; DT: pListe) und Inhalt (DT: inhaltT) |
    | Referenz:  um ein Element erweiterte Liste, aktualisierte Zeiger      |
    | Leistung:  Listenelement 'kreieren', der Variable neues^.daten den    |
    |            Wert der Variablen inhalt und dem Zeiger des neuen Ele-    |
    |            ments NIL zuweisen. Falls die Liste leer ist, wird das     |
    |            neue Element als erstes Element in die Liste bernommen,   |
    |            sonst am Ende der Liste angefgt.                          |
    *-----------------------------------------------------------------------*)

    VAR neues : pListe;

    BEGIN
      NEW(neues);
      neues^.daten:= inhalt;
      neues^.naechstes:= NIL;
      IF ListeLeer(liste) THEN                       (* Die Liste ist leer. *)
        liste.anker:= neues;
        liste.aktuelles:= neues;
        liste.letztes:= neues
      ELSE                     (* Die Liste enthlt mindestens ein Element. *)
        liste.letztes^.naechstes:= neues;
        liste.aktuelles:= neues;
        liste.letztes:= neues
      END
    END AnhaengenElement;


  (* 4. Operation *)

  PROCEDURE LoeschenElement* ( VAR liste : TListe );
(** Das aktuelle Element der Liste wird gelscht, die <liste> ist anschlieend
    um ein Element verkrzt. *)
(** Das aktuelle Element der Liste ist nach dieser Operation das erste, 
    sofern vorhanden. *)

   (*-----------------------------------------------------------------------*
    | bernahme: Verbund (drei Zeiger; DT: pListe)                          |
    | Referenz:  um ein Element verkrzte Liste, aktualisierte Zeiger       |
    | Leistung:  Lschen eines Listenelements und Umhngen der Zeiger.      |
    |            Einzelheiten werden im Quelltext kommentiert.              |
    *-----------------------------------------------------------------------*)

    VAR voraktuelles : pListe;

    BEGIN
      IF ~ListeLeer(liste) THEN 
              (* Die Liste enthlt  m i n d e s t e n s  e i n  Element.    *)
        IF liste.anker = liste.letztes THEN 
              (* Die Liste enthlt  g e n a u  e i n  Element.              *)
          liste.anker:= NIL
        ELSIF ListenAnfang(liste) THEN 
              (* Die Liste enthlt  m i n d e s t e n s  z w e i  Elemente. *)
              (* Das  e r s t e  Element ist  a k t u e l l e s  Element.   *)
          liste.anker:= liste.anker^.naechstes;
          liste.aktuelles:= liste.anker
        ELSIF ListenEnde(liste) THEN 
              (* Das  l e t z t e  Element ist  a k t u e l l e s  Element. *)
          voraktuelles:= Davor(liste);
          voraktuelles^.naechstes:= NIL;
          liste.letztes:= voraktuelles;
          liste.aktuelles:= liste.anker
        ELSE  (* W e d e r  das erste  n o c h  das letzte Element          *) 
              (* ist aktuelles Element.                                     *)
          voraktuelles:= Davor(liste);
          voraktuelles^.naechstes:= liste.aktuelles^.naechstes;
          liste.aktuelles:= liste.anker
        END
      END
    END LoeschenElement;


  (* 5. Operation *)

  PROCEDURE GeheErstes* ( VAR liste : TListe );
(** Verschiebt den Zeiger des aktuellen Elements auf das erste der Liste. *)
(** Sollte die Liste leer sein, passiert nichts (auch keine Fehlermeldung oder
    Programmabbruch). *)

   (*-----------------------------------------------------------------------*
    | bernahme: Verbund (drei Zeiger; DT: pListe)                          |
    | Referenz:  aktualisierte Zeiger (eigentlich nur: liste.aktuelles)     |
    | Leistung:  Falls die Liste nicht leer ist, wird der Zeiger            |
    |            liste.aktuelles auf den Listenanfang gesetzt.              |
    *-----------------------------------------------------------------------*)

    BEGIN
      IF ~ListeLeer(liste) THEN
        liste.aktuelles:= liste.anker
      END
    END GeheErstes;


  (* 6. Operation *)

  PROCEDURE GeheLetztes* ( VAR liste : TListe );
(** Verschiebt den Zeiger des aktuellen Elements auf das letzte der Liste. *)
(** Sollte die Liste leer sein, passiert nichts (auch keine Fehlermeldung oder
    Programmabbruch). *)

   (*-----------------------------------------------------------------------*
    | bernahme: Verbund (drei Zeiger; DT: pListe)                          |
    | Referenz:  aktualisierte Zeiger (eigentlich nur: liste.aktuelles)     |
    | Leistung:  Falls die Liste nicht leer ist, wird der Zeiger aktuelles  |
    |            auf das Listenende gesetzt.                                |
    *-----------------------------------------------------------------------*)

    BEGIN
      IF ~ListeLeer(liste) THEN
        liste.aktuelles:= liste.letztes
      END
    END GeheLetztes;


  (* 7. Operation *)

  PROCEDURE GeheNaechstes* ( VAR liste : TListe );
(** Verschiebt den Zeiger des aktuellen Elements eine Position 
    in der Liste weiter. *)
(** Wenn das aktuelle Element schon das letzte war, wird eine
    Fehlermeldung ausgegeben und das Programm abgebrochen. Der 
    Anwendungsprogrammierer muss dafr sorgen, dass dieser Fall nicht 
    eintritt! *)

   (*-----------------------------------------------------------------------*
    | bernahme: Verbund (drei Zeiger; DT: pListe)                          |
    | Referenz:  aktualisierte Zeiger (eigentlich nur: liste.aktuelles)     |
    | Leistung:  Falls die Liste nicht leer ist und der Zeiger aktuelles    |
    |            nicht auf das letzte Listenelement zeigt, wird der Zeiger  |
    |            aktuelles um eine Position in Richtung Listenende gesetzt. |
    |            Falls der Zeiger aktuelles bereits auf das letze Listen-   |
    |            element zeigt, wird eine Fehlermeldung ausgegeben und das  |
    |            Abarbeiten des Programms abgebrochen.                      |
    *-----------------------------------------------------------------------*)

    BEGIN
      IF ~ListeLeer(liste) THEN
        IF ListenEnde(liste) THEN
          D.GotoXY(2,24); 
          D.WriteStr('Laufzeitfehler bei GeheNaechstes:');
          D.GotoXY(2,25);  
          D.WriteStr('Zugriff nach dem letzten Element!');
          HALT(100)
        ELSE liste.aktuelles:= liste.aktuelles^.naechstes
        END
      END
    END GeheNaechstes;


  (* 8. Operation *)

  PROCEDURE HoleEintrag* ( liste : TListe; VAR inhalt : TInhalt );
(** Die Daten des aktuellen Elements werden in der Variablen <inhalt> 
    zurckgegeben. *)
(** Sollte die Liste leer sein, gibt es eine Fehlermeldung und das Programm
    wird abgebrochen. Der Anwendungsprogrammierer muss dafr sorgen, dass 
    dieser Fall nicht eintritt! *)

   (*-----------------------------------------------------------------------*
    | bernahme: Verbund (drei Zeiger; DT: pListe) und Inhalt (DT: inhaltT) |
    | Referenz:  inhalt (DT: inhaltT)                                       |
    | Leistung:  Ist die Liste leer, wird eine Fehlermeldung ausgegeben und |
    |            das Abarbeiten des Programms abgebrochen, sonst wird der   |
    |            Wert der Verbund-Variable liste.daten dem Referenzparame-  |
    |            ter inhalt zugewiesen.                                     |
    *-----------------------------------------------------------------------*)

    BEGIN
      IF ListeLeer(liste) THEN
        D.GotoXY(2,24); 
        D.WriteStr('Laufzeitfehler bei HoleEintrag:');
        D.GotoXY(2,25);  
        D.WriteStr('Liste ist leer!');
        HALT(101)
      ELSE inhalt:= liste.aktuelles^.daten
      END
    END HoleEintrag;


  (* 9. Operation *)

  PROCEDURE SchreibeEintrag* ( liste : TListe; inhalt : TInhalt );
(** Die Daten des aktuellen Elementes werden mit denen im Parameter 
    <inhalt> bergegebenen Daten berschrieben. *)
(** Sollte die Liste leer sein, gibt es eine Fehlermeldung und das Programm
    wird abgebrochen. Der Anwendungsprogrammierer muss dafr sorgen, dass 
    dieser Fall nicht eintritt! *)

   (*-----------------------------------------------------------------------*
    | bernahme: Verbund (drei Zeiger; DT: pListe) und Inhalt (DT: inhaltT) |
    | Referenz:  Liste, in der 'Inhalt' des Elements, auf das der Zeiger    |
    |            liste.aktuelles weist, berschrieben wurde                 |
    | Leistung:  Falls die Liste leer ist, wird eine Fehlermeldung ausgege- |
    |            ben und das Abarbeiten des Programms abgebrochen, sonst    |
    |            wird der Wert der Variable inhalt dem Datenfeld, auf das   |
    |            der Zeiger liste.aktuelles^ gerichtet ist, zugewiesen      |
    |            (liste.aktuelles^.daten:= inhalt).                         |
    *-----------------------------------------------------------------------*)

    BEGIN
      IF ListeLeer(liste) THEN
        D.GotoXY(2,24); 
        D.WriteStr('Laufzeitfehler bei SchreibeEintrag:');
        D.GotoXY(2,25);  
        D.WriteStr('Liste ist leer!');
        HALT(102)
      ELSE liste.aktuelles^.daten:= inhalt
      END
    END SchreibeEintrag;

 (* ------------------------  OHNE INITIALISIERUNG  ----------------------- *)

 END ADTListe.

 (* ------------------------------------------------------------------------*)
